==> audiochain-1.1.5-src.tgz: org/audiochain/devices/frequency/IntervalMagnitudeFftAudioDataReaderListener.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.lang.Math.*;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.*;

abstract public class IntervalMagnitudeFftAudioDataReaderListener implements FftAudioDataReaderListener {

	private final float frameRate;
	private final double maxAmplitude;
	private final double minDecibel;
	private static final double sqrtTwo = sqrt(2);
	private static final double oneDividedBySqrtTwo = 1 / sqrtTwo;
	private int intervalInMilliseconds = 60;
	private int numberOfFftsPerInterval;
	private int numberOfFftsCounter;
	private int rmsSpectrumLength;
	private double[] leftRmsMagnitudeSum;
	private double[] rightRmsMagnitudeSum;
	private int fftLength;

	public IntervalMagnitudeFftAudioDataReaderListener(float frameRate, int sampleSizeInBits) {
		this.frameRate = frameRate;
		this.maxAmplitude = (2 << sampleSizeInBits - 2) - 1;
		minDecibel = 20 * log10(1 / maxAmplitude);
		calcNumberOfFftsPerMeasurement();
	}

	@Override
	public final void fftReaderUpdate(double[][] leftSpectrum, double[][] rightSpectrum) {
		boolean omitRight = leftSpectrum == rightSpectrum;
		double[] leftMagnitude = hypot(normalizeSpectralDensity(realSpectrum(leftSpectrum)));
		double[] rightMagnitude = omitRight ? null : hypot(normalizeSpectralDensity(realSpectrum(rightSpectrum)));

		int length = leftMagnitude.length;
		if (leftRmsMagnitudeSum == null) {
			rmsSpectrumLength = length;
			leftRmsMagnitudeSum = new double[rmsSpectrumLength];
			if (!omitRight) {
				rightRmsMagnitudeSum = new double[rmsSpectrumLength];
			}
		}
		omitRight = omitRight || rightRmsMagnitudeSum == null;

		// only calculate the real part of the complex spectrum for performance
		// reasons
		int n = min(length, rmsSpectrumLength);
		for (int i = 0; i < n; i++) {
			// obtain the rms power of the magnitude
			leftRmsMagnitudeSum[i] += leftMagnitude[i] * oneDividedBySqrtTwo;
			if (!omitRight && rightMagnitude != null) { // nullcheck is just to suppress the compiler warning
				rightRmsMagnitudeSum[i] += rightMagnitude[i] * oneDividedBySqrtTwo;
			}
		}

		numberOfFftsCounter++;
		if (numberOfFftsPerInterval > 0 && numberOfFftsCounter % numberOfFftsPerInterval == 0) {
			numberOfFftsCounter = 0;
			for (int i = 0; i < n; i++) {
				double leftRmsSum = leftRmsMagnitudeSum[i];
				if (leftRmsSum == 0) {
					leftRmsMagnitudeSum[i] = 0;
				} else {
					double leftRms = leftRmsSum / numberOfFftsPerInterval;
					// obtain the db spl (sound pressure level in db) for max amplitude this equals -3dB
					double leftRmsDB = 20 * log10(leftRms / maxAmplitude);
					// scale to a value between 0 and 1
					leftRmsMagnitudeSum[i] = leftRmsDB > 0 ? 1D : (minDecibel - leftRmsDB) / minDecibel;
				}

				if (!omitRight) {
					double rightRmsSum = rightRmsMagnitudeSum[i];
					if (rightRmsSum == 0) {
						rightRmsMagnitudeSum[i] = 0;
					} else {
						double rightRms = rightRmsSum / numberOfFftsPerInterval;
						double rightRmsDB = 20 * log10(rightRms / maxAmplitude);
						rightRmsMagnitudeSum[i] = rightRmsDB > 0 ? 1D : (minDecibel - rightRmsDB) / minDecibel;
					}
				}
			}
			update(leftRmsMagnitudeSum, omitRight ? leftRmsMagnitudeSum : rightRmsMagnitudeSum);
			leftRmsMagnitudeSum = null;
			rightRmsMagnitudeSum = null;
		}
	}

	abstract protected void update(double[] leftRmsMagnitudeFactor, double[] rightRmsMagnitudeFactor);

	public int getIntervalInMilliseconds() {
		return intervalInMilliseconds;
	}

	public void setIntervalInMilliseconds(int interval) {
		this.intervalInMilliseconds = interval;
		calcNumberOfFftsPerMeasurement();
	}

	@Override
	public void fftLengthChanged(int oldFftLength, int newFftLength) {
		this.fftLength = newFftLength;
		calcNumberOfFftsPerMeasurement();
	}

	private void calcNumberOfFftsPerMeasurement() {
		if (fftLength > 0) {
			int numberOfValuesPerMeasurement = (int) (frameRate / (1000F / intervalInMilliseconds));
			numberOfFftsPerInterval = numberOfValuesPerMeasurement / fftLength;
		}
	}
}

==> audiochain-1.1.6-src.tgz: org/audiochain/devices/frequency/IntervalMagnitudeFftAudioDataReaderListener.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.frequency;

import static java.lang.Math.log10;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.hypot;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.normalizeSpectralDensity;
import static org.audiochain.devices.frequency.DigitalSignalProcessing.realSpectrum;

abstract public class IntervalMagnitudeFftAudioDataReaderListener implements FftAudioDataReaderListener {

	private final float frameRate;
	private final double maxAmplitude;
	private final double minDecibel;
	private static final double sqrtTwo = sqrt(2);
	private static final double oneDividedBySqrtTwo = 1 / sqrtTwo;
	private int intervalInMilliseconds = 60;
	private int numberOfFftsPerInterval;
	private int numberOfFftsCounter;
	private int rmsSpectrumLength;
	private double[] leftRmsMagnitudeSum;
	private double[] rightRmsMagnitudeSum;
	private int fftLength;

	public IntervalMagnitudeFftAudioDataReaderListener(float frameRate, int sampleSizeInBits) {
		this.frameRate = frameRate;
		this.maxAmplitude = (2 << sampleSizeInBits - 2) - 1;
		minDecibel = 20 * log10(1 / maxAmplitude);
		calcNumberOfFftsPerMeasurement();
	}

	@Override
	public final void fftReaderUpdate(double[][] leftSpectrum, double[][] rightSpectrum) {
		boolean omitRight = leftSpectrum == rightSpectrum;
		double[] leftMagnitude = hypot(normalizeSpectralDensity(realSpectrum(leftSpectrum)));
		double[] rightMagnitude = omitRight ? null : hypot(normalizeSpectralDensity(realSpectrum(rightSpectrum)));

		int length = leftMagnitude.length;
		if (leftRmsMagnitudeSum == null) {
			rmsSpectrumLength = length;
			leftRmsMagnitudeSum = new double[rmsSpectrumLength];
			if (!omitRight) {
				rightRmsMagnitudeSum = new double[rmsSpectrumLength];
			}
		}
		omitRight = omitRight || rightRmsMagnitudeSum == null;

		// only calculate the real part of the complex spectrum for performance
		// reasons
		int n = min(length, rmsSpectrumLength);
		for (int i = 0; i < n; i++) {
			// obtain the rms power of the magnitude
			leftRmsMagnitudeSum[i] += leftMagnitude[i] * oneDividedBySqrtTwo;
			if (!omitRight && rightMagnitude != null) { // nullcheck is just to suppress the compiler warning
				rightRmsMagnitudeSum[i] += rightMagnitude[i] * oneDividedBySqrtTwo;
			}
		}

		numberOfFftsCounter++;
		if (numberOfFftsPerInterval > 0 && numberOfFftsCounter % numberOfFftsPerInterval == 0) {
			numberOfFftsCounter = 0;
			for (int i = 0; i < n; i++) {
				double leftRmsSum = leftRmsMagnitudeSum[i];
				if (leftRmsSum == 0) {
					leftRmsMagnitudeSum[i] = 0;
				} else {
					double leftRms = leftRmsSum / numberOfFftsPerInterval;
					// obtain the db spl (sound pressure level in db) for max amplitude this equals -3dB
					double leftRmsDB = 20 * log10(leftRms / maxAmplitude);
					// scale to a value between 0 and 1
					leftRmsMagnitudeSum[i] = leftRmsDB > 0 ? 1D : (minDecibel - leftRmsDB) / minDecibel;
				}

				if (!omitRight) {
					double rightRmsSum = rightRmsMagnitudeSum[i];
					if (rightRmsSum == 0) {
						rightRmsMagnitudeSum[i] = 0;
					} else {
						double rightRms = rightRmsSum / numberOfFftsPerInterval;
						double rightRmsDB = 20 * log10(rightRms / maxAmplitude);
						rightRmsMagnitudeSum[i] = rightRmsDB > 0 ? 1D : (minDecibel - rightRmsDB) / minDecibel;
					}
				}
			}
			update(leftRmsMagnitudeSum, omitRight ? leftRmsMagnitudeSum : rightRmsMagnitudeSum);
			leftRmsMagnitudeSum = null;
			rightRmsMagnitudeSum = null;
		}
	}

	abstract protected void update(double[] leftRmsMagnitudeFactor, double[] rightRmsMagnitudeFactor);

	public int getIntervalInMilliseconds() {
		return intervalInMilliseconds;
	}

	public void setIntervalInMilliseconds(int interval) {
		this.intervalInMilliseconds = interval;
		calcNumberOfFftsPerMeasurement();
	}

	@Override
	public void fftLengthChanged(int oldFftLength, int newFftLength) {
		this.fftLength = newFftLength;
		calcNumberOfFftsPerMeasurement();
	}

	private void calcNumberOfFftsPerMeasurement() {
		if (fftLength > 0) {
			int numberOfValuesPerMeasurement = (int) (frameRate / (1000F / intervalInMilliseconds));
			numberOfFftsPerInterval = numberOfValuesPerMeasurement / fftLength;
		}
	}
}
